@@ -13,11 +13,12 @@ import multiprocessing as mp
 from multiprocessing import Queue
 
 from .utils.process_utils import complement_seq
+from .utils.process_utils import max_queue_size
 
 base = "C"
 pred_base = "CG"
 
-queue_size_border = 1000
+queue_size_border = max_queue_size
 time_wait = 1
 
 
@@ -364,17 +365,18 @@ def add_mm_ml_tags_to_bam(bamfile, per_readsite, modbamfile,
     p_w.join()
 
     if modbamfile.endswith(".bam"):
-        if mode == "align":
-            sys.stderr.write("sorting and indexing modbam file..\n")
-            modbam_sorted = modbamfile + ".sorted.bam"
+        try:
+            sys.stderr.write("sorting modbam file..\n")
+            modbam_sorted = os.path.splitext(modbamfile)[0] + ".sorted.bam"
             pysam.sort("-o", modbam_sorted, "-@", str(threads), modbamfile)
             os.rename(modbam_sorted, modbamfile)
-        else:
-            sys.stderr.write("indexing modbam file..\n")
+        except Exception:
+            sys.stderr.write("failed sorting modbam file..\n")
         try:
+            sys.stderr.write("indexing modbam file..\n")
             pysam.index("-@", str(threads), modbamfile)
         except Exception:
-            sys.stderr.write("failed on indexing modbam file..\n")
+            sys.stderr.write("failed indexing modbam file..\n")
 
     if os.path.exists(per_read_file):
         os.remove(per_read_file)
