@@ -183,8 +183,7 @@ def main():
                            help="retain at most n alignments in minimap2. "
                                 "default 3, which means 2 secondary alignments are retained. "
                                 "Do not use 2, cause -N1 is not suggested for high accuracy of alignment. "
-                                "[This arg is for further extension, for now it is no use cause "
-                                "we use only primary alignment.]")
+                                "[This arg is for further extension.]")
     sah_align.add_argument("--bwa", action="store_true", default=False, required=False,
                            help="use bwa instead of pbmm2 for alignment")
     sah_align.add_argument("--path_to_bwa", type=str, default=None, required=False,
@@ -208,11 +207,21 @@ def main():
     scm_input.add_argument("--holes_batch", type=int, default=50, required=False,
                            help="number of holes/hifi-reads in an batch to get/put in queues, default 50. "
                                 "only used when --input is bam/sam")
+    
+    scm_output = sub_call_mods.add_argument_group("OUTPUT")
+    scm_output.add_argument("--output", "-o", action="store", type=str, required=True,
+                            help="the prefix of output files to save the predicted results. "
+                                 "output files will be [--output].per_readsite.tsv/.modbam.bam")
+    scm_output.add_argument("--gzip", action="store_true", default=False, required=False,
+                            help="if compressing .per_readsite.tsv when --input is not in bam/sam format.")
+    scm_output.add_argument("--keep_pulse", action="store_true", default=False, required=False,
+                            help="if keeping ipd/pw tags in .modbam.bam when --input is in bam/sam format.")
+    scm_output.add_argument("--no_sort", action="store_true", default=False, required=False,
+                            help="don't sort .modbam.bam when --input is in bam/sam format.")
 
     scm_call = sub_call_mods.add_argument_group("CALL")
     scm_call.add_argument("--model_file", "-m", action="store", type=str, required=True,
                           help="file path of the trained model (.ckpt)")
-
     # model param
     scm_call.add_argument('--model_type', type=str, default="attbigru2s",
                           choices=["attbilstm2s", "attbigru2s"],
@@ -234,7 +243,6 @@ def main():
 
     scm_call.add_argument("--batch_size", "-b", default=512, type=int, required=False,
                           action="store", help="batch size, default 512")
-
     # BiRNN model param
     scm_call.add_argument('--n_vocab', type=int, default=16, required=False,
                           help="base_seq vocab_size (15 base kinds from iupac)")
@@ -243,23 +251,15 @@ def main():
     scm_call.add_argument('--layer_rnn', type=int, default=3,
                           required=False, help="BiRNN layer num, default 3")
     scm_call.add_argument('--hid_rnn', type=int, default=256, required=False,
-                          help="BiRNN hidden_size for combined feature")
-
-    scm_output = sub_call_mods.add_argument_group("OUTPUT")
-    scm_output.add_argument("--output", "-o", action="store", type=str, required=True,
-                            help="the prefix of output files to save the predicted results. "
-                                 "output files will be [--output].per_readsite.tsv/.modbam.bam")
-    scm_output.add_argument("--gzip", action="store_true", default=False, required=False,
-                            help="if compressing .per_readsite.tsv using gzip")
-    scm_output.add_argument("--modbam", type=str, default="yes", required=False,
-                            help="if generating modbam file when --input is in bam/sam format. "
-                                 "yes or no, default yes")
+                          help="BiRNN hidden_size, default 256")
 
     scm_extract = sub_call_mods.add_argument_group("EXTRACTION")
-    scm_extract.add_argument("--mode", type=str, default="align", required=False,
+    scm_extract.add_argument("--mode", type=str, default="denovo", required=False,
                              choices=["denovo", "align"],
-                             help="denovo mode: extract features from unaligned hifi.bam;\n"
-                                  "align mode: extract features from aligned hifi.bam. default: align")
+                             help="denovo mode: extract features from unaligned/aligned hifi.bam without "
+                                  "reference position info;\n"
+                                  "align mode: extract features from aligned hifi.bam with "
+                                  "reference position info. default: denovo")
     scm_extract.add_argument("--holeids_e", type=str, default=None, required=False,
                              help="file contains holeids to be extracted, default None")
     scm_extract.add_argument("--holeids_ne", type=str, default=None, required=False,
@@ -295,10 +295,10 @@ def main():
     scm_extract_ref = sub_call_mods.add_argument_group("EXTRACTION ALIGN_MODE")
     scm_extract_ref.add_argument("--ref", type=str, required=False,
                                  help="path to genome reference to be aligned, in fasta/fa format.")
-    scm_extract_ref.add_argument("--mapq", type=int, default=10, required=False,
-                                 help="MAPping Quality cutoff for selecting alignment items, default 10")
-    scm_extract_ref.add_argument("--identity", type=float, default=0.75, required=False,
-                                 help="identity cutoff for selecting alignment items, default 0.75")
+    scm_extract_ref.add_argument("--mapq", type=int, default=1, required=False,
+                                 help="MAPping Quality cutoff for selecting alignment items, default 1")
+    scm_extract_ref.add_argument("--identity", type=float, default=0.0, required=False,
+                                 help="identity cutoff for selecting alignment items, [0.0, 1.0], default 0.0")
     scm_extract_ref.add_argument("--no_supplementary", action="store_true", default=False, required=False,
                                  help="not use supplementary alignment")
     scm_extract_ref.add_argument("--is_mapfea", type=str, default="no", required=False,
@@ -334,10 +334,12 @@ def main():
                            help="if compressing the output using gzip")
 
     se_extract = sub_extract.add_argument_group("EXTRACTION")
-    se_extract.add_argument("--mode", type=str, default="align", required=False,
+    se_extract.add_argument("--mode", type=str, default="denovo", required=False,
                             choices=["denovo", "align"],
-                            help="denovo mode: extract features from unaligned hifi.bam;\n"
-                                 "align mode: extract features from aligned hifi.bam. default: align")
+                            help="denovo mode: extract features from unaligned/aligned hifi.bam without "
+                                 "reference position info;\n"
+                                 "align mode: extract features from aligned hifi.bam with "
+                                 "reference position info. default: denovo")
     se_extract.add_argument("--seq_len", type=int, default=21, required=False,
                             help="len of kmer. default 21")
     se_extract.add_argument("--motifs", action="store", type=str,
@@ -370,10 +372,10 @@ def main():
     se_extract_ref = sub_extract.add_argument_group("EXTRACTION ALIGN_MODE")
     se_extract_ref.add_argument("--ref", type=str, required=False,
                                 help="path to genome reference to be aligned, in fasta/fa format.")
-    se_extract_ref.add_argument("--mapq", type=int, default=10, required=False,
-                                help="MAPping Quality cutoff for selecting alignment items, default 10")
-    se_extract_ref.add_argument("--identity", type=float, default=0.75, required=False,
-                                help="identity cutoff for selecting alignment items, default 0.75")
+    se_extract_ref.add_argument("--mapq", type=int, default=1, required=False,
+                                help="MAPping Quality cutoff for selecting alignment items, default 1")
+    se_extract_ref.add_argument("--identity", type=float, default=0.0, required=False,
+                                help="identity cutoff for selecting alignment items, [0.0, 1.0], default 0.0")
     se_extract_ref.add_argument("--no_supplementary", action="store_true", default=False, required=False,
                                 help="not use supplementary alignment")
     se_extract_ref.add_argument("--is_mapfea", type=str, default="no", required=False,
@@ -477,18 +479,19 @@ def main():
     scfb_callfreq.add_argument('--call_mode', type=str, action="store", required=False, default="count",
                                choices=["count", "aggregate"],
                                help='call mode: count, aggregate. default count.')
-    scfb_callfreq.add_argument("--ag_model", type=str, action="store", required=False,
-                               help='model path for call_mode aggregate')
     scfb_callfreq.add_argument('--prob_cf', type=float, action="store", required=False, default=0.0,
-                               help='this is to remove ambiguous calls. '
+                               help='this is to remove ambiguous calls (only for count-mode now). '
                                     'if abs(prob1-prob0)>=prob_cf, then we use the call. e.g., proc_cf=0 '
                                     'means use all calls. range [0, 1], default 0.0.')
+    scfb_callfreq.add_argument('--no_amb_cov', action="store_true", required=False, default=False,
+                               help='when using prob_cf>0, DO NOT count ambiguous calls '
+                                    'for calculating reads coverage')
     scfb_callfreq.add_argument("--hap_tag", type=str, action="store", required=False, default="HP",
                                help="haplotype tag, default HP")
-    scfb_callfreq.add_argument("--mapq", type=int, default=10, required=False,
-                               help="MAPping Quality cutoff for selecting alignment items, default 10")
-    scfb_callfreq.add_argument("--identity", type=float, default=0.75, required=False,
-                               help="identity cutoff for selecting alignment items, default 0.75")
+    scfb_callfreq.add_argument("--mapq", type=int, default=1, required=False,
+                               help="MAPping Quality cutoff for selecting alignment items, default 1")
+    scfb_callfreq.add_argument("--identity", type=float, default=0.0, required=False,
+                               help="identity cutoff for selecting alignment items, [0.0, 1.0], default 0.0")
     scfb_callfreq.add_argument("--no_supplementary", action="store_true", default=False, required=False,
                                help="not use supplementary alignment")
     scfb_callfreq.add_argument("--motifs", action="store", type=str,
@@ -502,13 +505,44 @@ def main():
     scfb_callfreq.add_argument("--mod_loc", action="store", type=int, required=False, default=0,
                                help='0-based location of the targeted base in the motif, default 0')
     scfb_callfreq.add_argument("--no_comb", action="store_true", default=False, required=False,
-                               help="dont combine fwd/rev reads of one CG. [Only works when motifs is CG]")
+                               help="don't combine fwd/rev reads of one CG. [Only works when motifs is CG]")
     scfb_callfreq.add_argument('--refsites_only', action='store_true', default=False,
                                help="only keep sites which are target motifs in both reference and reads")
     scfb_callfreq.add_argument('--refsites_all', action='store_true', default=False,
                                help="output all covered sites which are target motifs in reference. "
                                     "--refsites_all is True, also means we do not output sites which "
                                     "are target motifs only in reads.")
+    scfb_callfreq.add_argument("--no_hap", action="store_true", default=False, required=False,
+                               help="don't call_freq on haplotypes ")
+    scfb_callfreq.add_argument("--base_clip", action="store", type=int, required=False, default=0,
+                               help='number of base clipped in each read, default 0')
+
+    scfb_aggre = sub_call_freqb.add_argument_group("AGGREGATE_MODE")
+    scfb_aggre.add_argument("--aggre_model", "-m", action="store", type=str, required=False,
+                            help="file path of the aggregate model (.ckpt)")
+    scfb_aggre.add_argument('--model_type', type=str, default="attbigru",
+                            choices=["attbilstm", "attbigru"],
+                            required=False,
+                            help="type of model to use, 'attbigru', 'attbilstm', "
+                                 "default: attbigru")
+    scfb_aggre.add_argument('--seq_len', type=int, default=11, required=False,
+                            help="len of sites used. default 11")
+    scfb_aggre.add_argument('--class_num', type=int, default=1, required=False)
+    scfb_aggre.add_argument('--layer_rnn', type=int, default=1,
+                            required=False, help="BiRNN layer num, default 1")
+    scfb_aggre.add_argument('--hid_rnn', type=int, default=32, required=False,
+                            help="BiRNN hidden_size, default 32")
+    scfb_aggre.add_argument('--bin_size', type=int, action="store", required=False, default=20,
+                            help="histogram bin size, default 20")
+    scfb_aggre.add_argument('--cov_cf', action="store", type=int, required=False,
+                            default=4, help="coverage cutoff, to consider if use aggregate model to "
+                                            "re-predict the modstate of the site")
+    scfb_aggre.add_argument('--only_close', action="store_true", default=False, required=False,
+                            help="[EXPERIMENTAL]")
+    scfb_aggre.add_argument('--discrete', action="store_true", default=False, required=False,
+                            help="[EXPERIMENTAL]")
+    scfb_aggre.add_argument('--tseed', type=int, default=1234,
+                            help='random seed for torch')
 
     sub_call_freqb.set_defaults(func=main_call_freqb)
 
@@ -548,14 +582,14 @@ def main():
     st_train.add_argument('--layer_rnn', type=int, default=3,
                           required=False, help="BiRNN layer num, default 3")
     st_train.add_argument('--hid_rnn', type=int, default=256, required=False,
-                          help="BiRNN hidden_size for combined feature")
+                          help="BiRNN hidden_size, default 256")
 
     st_training = sub_train.add_argument_group("TRAINING")
     # model training
     st_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
-                                                                                "Ranger"],
-                             required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
-                                                  "or 'Ranger', default Adam")
+                                                                                "Ranger", "LookaheadAdam"],
+                             required=False, help="type of optimizer to use, 'Adam', 'SGD', 'RMSprop', "
+                                                  "'Ranger' or 'LookaheadAdam', default Adam")
     st_training.add_argument('--batch_size', type=int, default=512, required=False)
     st_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
                              choices=["StepLR", "ReduceLROnPlateau"],
@@ -625,14 +659,14 @@ def main():
     stm_train.add_argument('--layer_rnn', type=int, default=3,
                            required=False, help="BiRNN layer num, default 3")
     stm_train.add_argument('--hid_rnn', type=int, default=256, required=False,
-                           help="BiRNN hidden_size for combined feature")
+                           help="BiRNN hidden_size, default 256")
 
     # model training
     stm_training = sub_trainm.add_argument_group("TRAINING")
     stm_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
-                                                                                 "Ranger"],
-                              required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
-                                                   "or 'Ranger', default Adam")
+                                                                                 "Ranger", "LookaheadAdam"],
+                              required=False, help="type of optimizer to use, 'Adam', 'SGD', 'RMSprop', "
+                                                   "'Ranger' or 'LookaheadAdam', default Adam")
     stm_training.add_argument('--batch_size', type=int, default=512, required=False)
     stm_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
                               choices=["StepLR", "ReduceLROnPlateau"],
@@ -727,7 +761,7 @@ def main():
     # sd_train.add_argument('--layer_rnn', type=int, default=3,
     #                       required=False, help="BiRNN layer num, default 3")
     # sd_train.add_argument('--hid_rnn', type=int, default=256, required=False,
-    #                       help="BiRNN hidden_size for combined feature")
+    #                       help="BiRNN hidden_size, default 256")
     #
     # sd_training = sub_denoise.add_argument_group("TRAINING")
     # # model training
