@@ -25,21 +25,21 @@ def parse_a_line(line):
 
     fkmer = np.array([base2code_dna[x] for x in words[5]])
     fpass = np.array([int(words[6])] * len(fkmer))
-    fipdm = np.array([float(x) for x in words[7].split(",")], dtype=np.float)
-    fipdsd = np.array([float(x) for x in words[8].split(",")], dtype=np.float) if words[8] != "." else 0
-    fpwm = np.array([float(x) for x in words[9].split(",")], dtype=np.float)
-    fpwsd = np.array([float(x) for x in words[10].split(",")], dtype=np.float) if words[10] != "." else 0
-    fqual = np.array([float(x) for x in words[11].split(",")], dtype=np.float)
-    fmap = np.array([float(x) for x in words[12].split(",")], dtype=np.float) if words[12] != "." else 0
+    fipdm = np.array([float(x) for x in words[7].split(",")], dtype=float)
+    fipdsd = np.array([float(x) for x in words[8].split(",")], dtype=float) if words[8] != "." else 0
+    fpwm = np.array([float(x) for x in words[9].split(",")], dtype=float)
+    fpwsd = np.array([float(x) for x in words[10].split(",")], dtype=float) if words[10] != "." else 0
+    fqual = np.array([float(x) for x in words[11].split(",")], dtype=float)
+    fmap = np.array([float(x) for x in words[12].split(",")], dtype=float) if words[12] != "." else 0
 
     rkmer = np.array([base2code_dna[x] for x in words[13]])
     rpass = np.array([int(words[14])] * len(rkmer))
-    ripdm = np.array([float(x) for x in words[15].split(",")], dtype=np.float)
-    ripdsd = np.array([float(x) for x in words[16].split(",")], dtype=np.float) if words[16] != "." else 0
-    rpwm = np.array([float(x) for x in words[17].split(",")], dtype=np.float)
-    rpwsd = np.array([float(x) for x in words[18].split(",")], dtype=np.float) if words[18] != "." else 0
-    rqual = np.array([float(x) for x in words[19].split(",")], dtype=np.float)
-    rmap = np.array([float(x) for x in words[20].split(",")], dtype=np.float) if words[20] != "." else 0
+    ripdm = np.array([float(x) for x in words[15].split(",")], dtype=float)
+    ripdsd = np.array([float(x) for x in words[16].split(",")], dtype=float) if words[16] != "." else 0
+    rpwm = np.array([float(x) for x in words[17].split(",")], dtype=float)
+    rpwsd = np.array([float(x) for x in words[18].split(",")], dtype=float) if words[18] != "." else 0
+    rqual = np.array([float(x) for x in words[19].split(",")], dtype=float)
+    rmap = np.array([float(x) for x in words[20].split(",")], dtype=float) if words[20] != "." else 0
 
     label = int(words[21])
 
@@ -147,3 +147,48 @@ class FeaData3(Dataset):
 
     def close(self):
         pass
+
+
+# Dataloader for aggregate mode ==================================================================
+def parse_a_line_aggre(line):
+    # chrom, pos, strand, [offset_location_from_centerpos], [modprobs_histogram], [coverage], label
+    words = line.strip().split("\t")
+
+    sampleinfo = "\t".join(words[0:3])
+
+    offsets = np.array([int(x) for x in words[3].split(",")])
+    histos = np.array([[float(y) for y in x.split(",")] for x in words[4].split(";")])
+    covs = np.array([int(x) for x in words[5].split(",")])
+
+    label = int(words[6])
+
+    return sampleinfo, offsets, histos, covs, label
+
+
+class AggreFeaData(Dataset):
+    def __init__(self, filename, transform=None):
+        # print(">>>using linecache to access '{}'<<<\n"
+        #       ">>>after done using the file, "
+        #       "remember to use linecache.clearcache() to clear cache for safety<<<".format(filename))
+        self._filename = os.path.abspath(filename)
+        self._total_data = 0
+        self._transform = transform
+        # self.max_subreads = max_subreads
+        with open(filename, "r") as f:
+            self._total_data = len(f.readlines())
+
+    def __getitem__(self, idx):
+        line = linecache.getline(self._filename, idx + 1)
+        if line == "":
+            return None
+        else:
+            output = parse_a_line_aggre(line)
+            if self._transform is not None:
+                output = self._transform(output)
+            return output
+
+    def __len__(self):
+        return self._total_data
+
+    def close(self):
+        pass
