@@ -45,14 +45,14 @@ class ModelAttRNN(nn.Module):
             self.rnn_cell = "lstm"
             self.rnn = nn.LSTM(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
                                dropout=dropout_rate, batch_first=True, bidirectional=True)
-            self.rnn2 = nn.LSTM(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
-                                dropout=dropout_rate, batch_first=True, bidirectional=True)
+            # self.rnn2 = nn.LSTM(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
+            #                     dropout=dropout_rate, batch_first=True, bidirectional=True)
         elif self.model_type == "attbigru2s":
             self.rnn_cell = "gru"
             self.rnn = nn.GRU(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
                               dropout=dropout_rate, batch_first=True, bidirectional=True)
-            self.rnn2 = nn.GRU(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
-                               dropout=dropout_rate, batch_first=True, bidirectional=True)
+            # self.rnn2 = nn.GRU(embedding_size + self.feas_ccs, self.hidden_size, self.num_layers,
+            #                    dropout=dropout_rate, batch_first=True, bidirectional=True)
         else:
             raise ValueError("--model_type not set right!")
 
@@ -60,7 +60,7 @@ class ModelAttRNN(nn.Module):
         self.fc1 = nn.Linear(self.hidden_size * 2 * 2, self.num_classes)  # 2 for bidirection, another 2 for 2 strands
 
         self._att3 = Attention(self.hidden_size * 2, self.hidden_size * 2, self.hidden_size)
-        self._att3_2 = Attention(self.hidden_size * 2, self.hidden_size * 2, self.hidden_size)
+        # self._att3_2 = Attention(self.hidden_size * 2, self.hidden_size * 2, self.hidden_size)
 
         self.softmax = nn.Softmax(1)
 
@@ -118,6 +118,9 @@ class ModelAttRNN(nn.Module):
         out1, n_states1 = self.rnn(out1, self.init_hidden(out1.size(0),
                                                           self.num_layers,
                                                           self.hidden_size))  # (N, L, nhid*2)
+        # out2, n_states2 = self.rnn2(out2, self.init_hidden(out2.size(0),
+        #                                                    self.num_layers,
+        #                                                    self.hidden_size))  # (N, L, nhid*2)
         out2, n_states2 = self.rnn(out2, self.init_hidden(out2.size(0),
                                                           self.num_layers,
                                                           self.hidden_size))  # (N, L, nhid*2)
@@ -133,7 +136,8 @@ class ModelAttRNN(nn.Module):
         h_n2 = n_states2[0] if self.rnn_cell == "lstm" else n_states2
         h_n2 = h_n2.reshape(self.num_layers, 2, -1, self.hidden_size)[-1]  # last layer (2, N, nhid)
         h_n2 = h_n2.transpose(0, 1).reshape(-1, 1, 2 * self.hidden_size)
-        out2, att_weights2 = self._att3_2(h_n2, out2)
+        # out2, att_weights2 = self._att3_2(h_n2, out2)
+        out2, att_weights2 = self._att3(h_n2, out2)
 
         out = torch.cat((out1, out2), 1)
 
@@ -143,10 +147,11 @@ class ModelAttRNN(nn.Module):
         return out, self.softmax(out)
 
 
+# Using BS-seq modfreqs as gold standard to train a AggrAttRNN regression model, no softmax
 class AggrAttRNN(nn.Module):
-    def __init__(self, seq_len=11, num_layers=1, num_classes=2,
-                 dropout_rate=0.5, hidden_size=256,
-                 model_type="attbilstm",
+    def __init__(self, seq_len=11, num_layers=1, num_classes=1,
+                 dropout_rate=0.5, hidden_size=32, binsize=20,
+                 model_type="attbigru",
                  device=0):
         super(AggrAttRNN, self).__init__()
         self.model_type = model_type
@@ -157,25 +162,24 @@ class AggrAttRNN(nn.Module):
         self.num_classes = num_classes
         self.hidden_size = hidden_size
 
-        self.feas_ccs = 20 + 1
+        self.feas_ccs = binsize + 1
         if self.model_type == "attbilstm":
             self.rnn_cell = "lstm"
             self.rnn = nn.LSTM(self.feas_ccs, self.hidden_size, self.num_layers,
-                               dropout=dropout_rate, batch_first=True, bidirectional=True)
+                               dropout=0, batch_first=True, bidirectional=True)
         elif self.model_type == "attbigru":
             self.rnn_cell = "gru"
             self.rnn = nn.GRU(self.feas_ccs, self.hidden_size, self.num_layers,
-                              dropout=dropout_rate, batch_first=True, bidirectional=True)
+                              dropout=0, batch_first=True, bidirectional=True)
         else:
             raise ValueError("--model_type not set right!")
 
-        self.dropout1 = nn.Dropout(p=dropout_rate)
-        self.fc1 = nn.Linear(self.hidden_size * 2 * 2, self.num_classes)  # 2 for bidirection, another 2 for 2 strands
-
         self._att3 = Attention(self.hidden_size * 2, self.hidden_size * 2, self.hidden_size)
-        self._att3_2 = Attention(self.hidden_size * 2, self.hidden_size * 2, self.hidden_size)
 
-        self.softmax = nn.Softmax(1)
+        self.dropout1 = nn.Dropout(p=dropout_rate)
+        self.fc1 = nn.Linear(self.hidden_size * 2, self.num_classes)  # 2 for bidirection
+
+        # self.softmax = nn.Softmax(1)
 
     def get_model_type(self):
         return self.model_type
@@ -183,17 +187,20 @@ class AggrAttRNN(nn.Module):
     def init_hidden(self, batch_size, num_layers, hidden_size):
         # Set initial states
         h0 = torch.randn(num_layers * 2, batch_size, hidden_size, requires_grad=True)
-        if use_cuda:
+        if use_cuda and self.device != "cpu":
             h0 = h0.cuda(self.device)
         if self.rnn_cell == "lstm":
             c0 = torch.randn(num_layers * 2, batch_size, hidden_size, requires_grad=True)
-            if use_cuda:
+            if use_cuda and self.device != "cpu":
                 c0 = c0.cuda(self.device)
             return h0, c0
         return h0
 
-    def forward(self, feaseq):
-        out = torch.reshape(feaseq, (-1, self.seq_len, 1)).float()
+    def forward(self, offsets, histos):
+
+        offsets = torch.reshape(offsets, (-1, self.seq_len, 1)).float()  # (N, L, 1)
+
+        out = torch.cat((histos.float(), offsets), 2)
 
         out, n_states = self.rnn(out, self.init_hidden(out.size(0),
                                                        self.num_layers,
@@ -208,5 +215,6 @@ class AggrAttRNN(nn.Module):
 
         out = self.dropout1(out)
         out = self.fc1(out)
+        # out = self.softmax(out)
 
-        return out, self.softmax(out)
+        return out
