@@ -125,6 +125,12 @@ def train(args):
         except ImportError:
             raise ImportError("please check if ranger2020.py is in utils/ dir!")
         optimizer = Ranger(model.parameters(), lr=args.lr, betas=(0.95, 0.999), eps=1e-5)
+    elif args.optim_type == "LookaheadAdam":
+        try:
+            from .utils.lookahead import LookaheadAdam
+        except ImportError:
+            raise ImportError("please check if lookahead.py is in utils/ dir!")
+        optimizer = LookaheadAdam(model.parameters(), lr=args.lr)
     else:
         raise ValueError("--optim_type is not right!")
 
@@ -347,14 +353,14 @@ def main():
     st_train.add_argument('--layer_rnn', type=int, default=3,
                           required=False, help="BiRNN layer num, default 3")
     st_train.add_argument('--hid_rnn', type=int, default=256, required=False,
-                          help="BiRNN hidden_size for combined feature")
+                          help="BiRNN hidden_size, default 256")
 
     st_training = parser.add_argument_group("TRAINING")
     # model training
     st_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
-                                                                                "Ranger"],
-                             required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
-                                                  "or 'Ranger', default Adam")
+                                                                                "Ranger", "LookaheadAdam"],
+                             required=False, help="type of optimizer to use, 'Adam', 'SGD', 'RMSprop', "
+                                                  "'Ranger' or 'LookaheadAdam', default Adam")
     st_training.add_argument('--batch_size', type=int, default=512, required=False)
     st_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
                              choices=["StepLR", "ReduceLROnPlateau"],
