@@ -133,7 +133,7 @@ def train_worker(local_rank, global_world_size, args):
     model = model.cuda(local_rank)
     # DistributedDataParallel
     model = DDP(model, device_ids=[local_rank], output_device=local_rank,
-                find_unused_parameters=True)
+                find_unused_parameters=False)
 
     # 2. define dataloader
     print("training_process-{} reading data..".format(os.getpid()))
@@ -183,6 +183,12 @@ def train_worker(local_rank, global_world_size, args):
         except ImportError:
             raise ImportError("please check if ranger2020.py is in utils/ dir!")
         optimizer = Ranger(model.parameters(), lr=args.lr, betas=(0.95, 0.999), eps=1e-5)
+    elif args.optim_type == "LookaheadAdam":
+        try:
+            from .utils.lookahead import LookaheadAdam
+        except ImportError:
+            raise ImportError("please check if lookahead.py is in utils/ dir!")
+        optimizer = LookaheadAdam(model.parameters(), lr=args.lr)
     else:
         raise ValueError("--optim_type is not right!")
 
@@ -214,7 +220,7 @@ def train_worker(local_rank, global_world_size, args):
                 _, fkmer, fpass, fipdm, fipdsd, fpwm, fpwsd, fqual, fmap, \
                     rkmer, rpass, ripdm, ripdsd, rpwm, rpwsd, rqual, rmap, \
                     labels = sfeatures
-                # TODO: on_blocking=True or False?
+                # TODO: non_blocking=True or False?
                 fkmer = fkmer.cuda(local_rank, non_blocking=True)
                 fpass = fpass.cuda(local_rank, non_blocking=True)
                 fipdm = fipdm.cuda(local_rank, non_blocking=True)
@@ -452,14 +458,14 @@ def main():
     st_train.add_argument('--layer_rnn', type=int, default=3,
                           required=False, help="BiRNN layer num, default 3")
     st_train.add_argument('--hid_rnn', type=int, default=256, required=False,
-                          help="BiRNN hidden_size for combined feature")
+                          help="BiRNN hidden_size, default 256")
 
     st_training = parser.add_argument_group("TRAINING")
     # model training
     st_training.add_argument('--optim_type', type=str, default="Adam", choices=["Adam", "RMSprop", "SGD",
-                                                                                "Ranger"],
-                             required=False, help="type of optimizer to use, 'Adam' or 'SGD' or 'RMSprop' "
-                                                  "or 'Ranger', default Adam")
+                                                                                "Ranger", "LookaheadAdam"],
+                             required=False, help="type of optimizer to use, 'Adam', 'SGD', 'RMSprop', "
+                                                  "'Ranger' or 'LookaheadAdam', default Adam")
     st_training.add_argument('--batch_size', type=int, default=512, required=False)
     st_training.add_argument('--lr_scheduler', type=str, default='StepLR', required=False,
                              choices=["StepLR", "ReduceLROnPlateau"],
